/*
 * safenet.c - model specific routines for following units:
 *
 * - Fairstone L525/-625/-750
 * - Fenton P400/-600/-800
 * - Gemini  UPS625/-1000
 * - Powerwell PM525A/-625A/-800A/-1000A/-1250A
 * - Repotec RPF525/-625/-800/-1000
 * - Soltec Winmate 525/625/800/1000
 * - Sweex 500/1000 *)
 * - others using SafeNet software and serial interface
 *
 *   *) tested
 * 
 * Compliments to Sweex for claiming UNIX compatibility in their
 * instruction manual. "This UPS is compatible with UNIX, because it
 * can deliver power your computer when the mains fails". Yeah right,
 * that's why I pay a premium for a manageable UPS (it was still cheap
 * though). If you use this driver, it's because of them that I had to
 * write it so you may wish to thank them if you find it useful.
 *
 * Status:
 *  20031015/Revision 0.1 - Arjen de Korte <arjen@de-korte.org>
 *   - initial release (entirely based on reverse engineering the
 *     serial data stream)
 *
 * Copyright (C) 2003  Arjen de Korte <arjen@de-korte.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <ctype.h>
#include <sys/ioctl.h>

#include "main.h"
#include "serial.h"
#include "safenet.h"

#define DRV_VERSION     "0.01"

#define ENDCHAR         '\r'
#define IGNCHARS        ""
#define	UPSDELAY	10000

/*
 * Here we keep the last known status of the UPS
 */
static safenet_u	ups;

static int safenet_command(char *command)
{
	char	reply[256];
	int	i, ret;

	upsdebugx(2, "-> safenet_command");

	/*
	 * The command we send here is just one of the seemingly
	 * random commands thrown at the UPS by the SafeNet software.
	 * It doesn't seem to care which one, although totally random
	 * data is not accepted. Hmmm, wonder what the meaning is.
	 */

	ser_flush_in(upsfd, IGNCHARS, 0);
	ser_send_pace(upsfd, UPSDELAY, command);
	ret = ser_get_line(upsfd, reply, sizeof(reply), ENDCHAR, IGNCHARS, 1, 0);

	upsdebugx(1, "   %s", ((strlen(reply) > 0) ? reply : "empty"));

	if ((ret != 11) && (reply[0] != '$') && (strspn(reply+1, "AB") != 10))
	{
		dstate_datastale();
		upslogx(LOG_ERR, "Status read failed");
		return(0);
	}
	
	for (i=0; i<10; i++)
	{
		ups.reply[i] = ((reply[i+1] == 'A') ? FALSE : TRUE);
	}

	status_init();

	if (ups.status.onbattery)
		status_set("OB");
	else
		status_set("OL");
		
	if (ups.status.batterylow)
		status_set("BL");

	if (ups.status.overload)
		status_set("OVER");

	if (ups.status.batteryfail)
		status_set("RB");

	if (ups.status.systemtest)
		status_set("CAL");

	status_commit();

        dstate_dataok();

	return(1);
}

static void shutdown_stayoff(int delay)
{
	char command[16];

	upsdebugx(2, "-> shutdown_stayoff");

	if ((delay < 1) || (delay > 999))
	{
		upslogx(LOG_ERR, "Shutdown delay %d is not within valid range [1..999]");
		return;
	}

	snprintf(command, sizeof(command), "ZBAS%c%c%cWLPGE\r", ((delay / 100) + 'A'),
		(((delay % 100) / 10) + 'A'), ((delay % 10) + 'A'));

	upsdebugx(1, "Shutdown command : %s", command);

	/*
	 * The shutdown command is a one way street. After reception of this command,
	 * the UPS no longer accepts commands and will not reply with a status word.
	 */
	safenet_command(command);

	return;
}

static void shutdown_restart(int delay, int restart)
{
	char command[16];

	upsdebugx(2, "-> shutdown_restart");

	/*
	 * Bounds checking to prevent bogus commands
	 */
	if ((delay < 1) || (delay > 999))
	{
		upslogx(LOG_ERR, "Shutdown delay %d is not within valid range [1..999]", delay);
		return;
	}

	if ((restart < 1) || (restart > 9999))
	{
		upslogx(LOG_ERR, "Restart delay %d is not within valid range [1..9999]", restart);
		return;
	}

	snprintf(command, sizeof(command), "ZAF%c%c%cR%c%c%c%cO\r", ((delay / 100) + 'A'),
		(((delay % 100) / 10) + 'A'), ((delay % 10) + 'A'),
		((restart / 1000) + 'A'), (((restart % 1000) / 100) + 'A'),
		(((restart % 100) / 10) + 'A'), ((restart % 10) + 'A'));

	upsdebugx(1, "Shutdown command : %s", command);

	/*
	 * The shutdown command is a one way street. After reception of this command,
	 * the UPS no longer accepts commands and will not reply with a status word.
	 */
	safenet_command(command);

	return;
}

static int instcmd(const char *cmdname, const char *extra)
{
	upsdebugx(2, "-> instcmd");

	if (!strcasecmp(cmdname, "test.battery.start"))
	{
		safenet_command(SN_TBATT1);
		return STAT_INSTCMD_HANDLED;
	}

	if (!strcasecmp(cmdname, "test.battery.stop"))
	{
		safenet_command(SN_TBATT0);
		return STAT_INSTCMD_HANDLED;
	}

	if (!strcasecmp (cmdname, "test.failure.start"))
	{
		safenet_command(SN_TFAIL1);
		return STAT_INSTCMD_HANDLED;
	}

	if (!strcasecmp (cmdname, "test.failure.stop"))
	{
		safenet_command(SN_TFAIL0);
		return STAT_INSTCMD_HANDLED;
	}

	if (!strcasecmp(cmdname, "beeper.on"))
	{ 
		if (!ups.status.beeper) safenet_command(SN_BEEPER);
		return STAT_INSTCMD_HANDLED;
	}

	if (!strcasecmp(cmdname, "beeper.off"))
	{ 
		if (ups.status.beeper) safenet_command(SN_BEEPER);
		return STAT_INSTCMD_HANDLED;
	}

	if (!strcasecmp(cmdname, "shutdown.reboot"))
	{ 
		shutdown_restart(1, 1);
		return STAT_INSTCMD_HANDLED;
	}

	if (!strcasecmp(cmdname, "shutdown.reboot.graceful"))
	{ 
		shutdown_restart(120, 1);
		return STAT_INSTCMD_HANDLED;
	}

	upslogx(LOG_NOTICE, "instcmd: unknown command [%s]", cmdname);
	return STAT_INSTCMD_UNKNOWN;
}

void upsdrv_initinfo(void)
{
	int	retry = 0;

	upsdebugx(2, "-> upsdrv_initinfo");

	dstate_setinfo("driver.version.internal", "%s", DRV_VERSION);

	/*
	 * Initialize the serial interface of the UPS by sending the magic
	 * string. If it does not respond with a valid status reply within
	 * five tries, display error message and give up.
	 */
	while (!safenet_command(SN_INIT))
	{
		if (++retry > 4)
			fatalx("SafeNet protocol compatible UPS not found on %s", device_path);
	}

	dstate_addcmd ("test.battery.start");
	dstate_addcmd ("test.battery.stop");
	dstate_addcmd ("test.failure.start");
	dstate_addcmd ("test.failure.stop");
	dstate_addcmd ("shutdown.reboot");
	dstate_addcmd ("shutdown.reboot.graceful");

	upsh.instcmd = instcmd;
}

void upsdrv_updateinfo(void)
{
	upsdebugx(2, "-> upsdrv_updateinfo");

	/*
	 * The command we send here is just one of the seemingly
	 * random commands thrown at the UPS by the SafeNet software.
	 * It doesn't seem to care which one, although totally random
	 * data is not accepted. Hmmm, wonder what the meaning is.
	 */
	safenet_command(SN_STATUS);
}

void upsdrv_shutdown(void)
{
	int	retry = 0;

	upsdebugx(2, "-> upsdrv_shutdown");

	/*
	 * Since we may have arrived here before the hardware is initialized,
	 * try to initialize it here three times. If it fails, go on anyway
	 * since we may be running on batteries now and don't want to waste
	 * too much time trying.
	 */
	while (!safenet_command(SN_INIT))
	{
		if (retry++ > 2) break;
	}

	/*
	 * Since the UPS will happily restart on battery, we must use a
	 * different shutdown command depending on the line status, so
	 * we need to check the status of the UPS here.
	 */
	if (ups.status.onbattery)
	{
		/*
                 * Kill the inverter and wait for the things to
		 * come. Time to shutdown is 10 seconds.
		 */
		upslogx(LOG_NOTICE, "Shutdown and stay off");
		shutdown_stayoff(10);
	}
	else
	{
		/*
		 * Apparently we're not running on battery right
		 * now. Let's cycle the UPS and try to prevent a
		 * lockup. Time to shutdown is 10 seconds, after
		 * 1 minute the power returns.
		 */
		upslogx(LOG_NOTICE, "Shutdown and restart");
		shutdown_restart(10, 1);
	}
}

void upsdrv_help(void)
{
	upsdebugx(2, "-> upsdrv_help");
}

void upsdrv_makevartable(void)
{
	upsdebugx(2, "-> upsdrv_makevartable");
}

void upsdrv_banner(void)
{
	upsdebugx(2, "-> upsdrv_banner");

	printf("Network UPS Tools - Generic SafeNet UPS driver %s (%s)\n\n", DRV_VERSION, UPS_VERSION);
	experimental_driver = 1;
}

void upsdrv_initups(void)
{
	int dtr_bit = TIOCM_DTR;
	int rts_bit = TIOCM_RTS;

	upsdebugx(2, "-> upsdrv_initups");

	/*
	 * Open and lock the serial port and set the speed to 1200 baud.
	 */
	upsfd = ser_open(device_path);
	ser_set_speed(upsfd, device_path, B1200);

	/*
	 * Set DTR and clear RTS to provide power for the serial interface.
	 */
	ioctl(upsfd, TIOCMBIS, &dtr_bit);
	ioctl(upsfd, TIOCMBIC, &rts_bit);

	/*
	 * Wait one second to allow the voltages to settle before going on.
	 */
	sleep(1);
}

void upsdrv_cleanup(void)
{
	upsdebugx(2, "-> upsdrv_cleanup");

	ser_close(upsfd, device_path);
}
